import os
import yaml
import logging
import subprocess

from .recipe import Recipe, RecipeError
from .worktree import RecipeInWorktree


logger = logging.getLogger(__name__)


class _Skip(RuntimeError):
    pass


class _Failure(RuntimeError):
    pass


class _TestFailed(_Failure):
    def __init__(self, output):
        super().__init__("test failed")
        self.output = output


def yn_question(question, default):
    default_txt = "[Y/n]" if default else "[y/N]"
    while True:
        txt = input(f"{question} {default_txt} ").strip().lower()
        if not txt:
            return default
        elif txt[0] == "y":
            return True
        elif txt[0] == "n":
            return False


def branch_exists(recipe, branch_name):
    return (
        subprocess.call(
            ["git", "show-ref", "--verify", "-q", f"refs/heads/{branch_name}"],
            cwd=recipe.path,
        )
        == 0
    )


def create_branch_and_commit(recipe, branch_name, commit_msg):
    subprocess.check_call(["git", "checkout", "-q", "-b", branch_name], cwd=recipe.path)
    subprocess.check_call(
        [
            "git",
            "commit",
            "-q",
            "-a",
            "-m",
            commit_msg,
        ],
        cwd=recipe.path,
    )


def remove_branch(recipe, branch_name):
    subprocess.check_call(["git", "branch", "-q", "-D", branch_name], cwd=recipe.path)


def push_branch(recipe, remote, branch_name, force):
    subprocess.check_output(
        ["git", "push", "-q", "--set-upstream"]
        + (["-f"] if force else [])
        + [remote, branch_name],
        cwd=recipe.path,
    )


def add_version(recipe, folder, conan_version, upstream_version):
    conandata_path = os.path.join(recipe.path, folder, "conandata.yml")
    if not os.path.exists(conandata_path):
        raise _Failure("no conandata.yml")

    with open(conandata_path) as fil:
        conandata = yaml.load(fil, Loader=yaml.FullLoader)

    url = recipe.upstream.source_url(upstream_version)
    hash_digest = recipe.upstream.source_sha256_digest(upstream_version)
    config = recipe.config()
    config["versions"][conan_version] = {"folder": folder}
    conandata["sources"][conan_version] = {"sha256": hash_digest, "url": url}

    with open(recipe.config_file_path, "w") as fil:
        yaml.dump(config, fil)

    with open(conandata_path, "w") as fil:
        yaml.dump(conandata, fil)


def test_recipe(recipe, folder, version_str):
    version_folder_path = os.path.join(recipe.path, folder)

    env = os.environ.copy()
    env["CONAN_HOOK_ERROR_LEVEL"] = "40"

    if not logger.isEnabledFor(logging.DEBUG):
        stdout = stderr = subprocess.DEVNULL
    else:
        stdout = stderr = None

    code = subprocess.call(
        ["conan", "create", ".", f"{recipe.name}/{version_str}@"],
        env=env,
        stdout=stdout,
        stderr=stderr,
        cwd=version_folder_path,
    )

    if code != 0:
        raise _Failure("test failed")


def _get_most_recent_upstream_version(recipe):
    status = recipe.status()

    if status.up_to_date():
        raise _Skip("recipe is up-to-date")

    if not status.update_possible():
        raise _Failure("update is not possible")

    upstream_version = status.upstream_version
    if upstream_version.unknown:
        raise _Failure("upstream version is unknown")
    return upstream_version


def _get_user_choice_upstream_version(recipe):
    recipe_versions_fixed = [v.fixed for v in recipe.versions]
    versions = list(
        sorted(
            v for v in recipe.upstream.versions if v.fixed not in recipe_versions_fixed
        )
    )
    if not versions:
        raise _Failure("update is not possible")

    print("Choose an upstream version:")
    for i, v in enumerate(versions):
        print(f"{i:3d}) {v}")

    upstream_version = None
    while upstream_version is None:
        try:
            upstream_version = versions[int(input("Choice: "))]
        except (ValueError, KeyError):
            pass
    return upstream_version


def update_one_recipe(
    cci_path, recipe_name, choose_version, folder, run_test, push, force
):
    recipe = Recipe(cci_path, recipe_name)
    if getattr(recipe.conanfile_class, "deprecated", False):
        raise _Skip("recipe is deprecated")

    if choose_version:
        upstream_version = _get_user_choice_upstream_version(recipe)
    else:
        upstream_version = _get_most_recent_upstream_version(recipe)

    conan_version = upstream_version.fixed
    branch_name = f"{recipe.name}-{conan_version}"
    if branch_exists(recipe, branch_name):
        if not force:
            force = yn_question(
                f"Branch '{branch_name}' already exists, overwrite ?", False
            )
        if not force:
            raise _Failure(f"branch '{branch_name}' already exists")
        remove_branch(recipe, branch_name)

    folder = folder or recipe.versions_folders[recipe.most_recent_version]

    logger.info(
        "%s: adding version %s to folder %s in branch %s from upstream version %s",
        recipe_name,
        conan_version,
        folder,
        branch_name,
        upstream_version,
    )

    with RecipeInWorktree(recipe) as new_recipe:
        add_version(new_recipe, folder, conan_version, upstream_version)

        if run_test:
            logger.info("%s: running test", recipe_name)
            test_recipe(new_recipe, folder, conan_version)

        create_branch_and_commit(
            new_recipe, branch_name, f"{recipe.name}: add version {conan_version}"
        )

        if push:
            logger.info("%s: pushing", recipe_name)
            push_branch(new_recipe, "origin", branch_name, force)

    logger.info(
        "%s: created version %s in branch %s (%s)",
        recipe_name,
        conan_version,
        branch_name,
        "pushed" if push else "not pushed",
    )


def update_recipes(cci_path, recipes, choose_version, folder, run_test, push, force):
    ok = True
    for recipe in recipes:
        try:
            update_one_recipe(
                cci_path, recipe, choose_version, folder, run_test, push, force
            )
        except _Skip:
            logger.info("%s: skipped (%s)", recipe, str(exc))
        except (_Failure, RecipeError) as exc:
            logger.error("%s: %s", recipe, str(exc))
            ok = False

    return 0 if ok else 1



import re
import json
import time
import logging
import datetime
from multiprocessing import Pool

import requests
from terminaltables import AsciiTable
from colored import fg, stylize

from .recipe import get_recipes_list, Recipe
from .version import Version
from .github import get_github_token

ISSUE_URL_RE = re.compile(r"github.com/([^/]+)/([^/]+)/issues/([0-9]+)")
logger = logging.getLogger(__name__)


def status_one_recipe(cci_path, recipe_name):
    return Recipe(cci_path, recipe_name).status()


def get_status(cci_path, recipes, jobs):
    logger.info(f"Parsing {len(recipes)} recipes...")

    with Pool(jobs) as p:
        status_futures = [
            p.apply_async(status_one_recipe, args=(cci_path, recipe))
            for recipe in recipes
        ]
        return [f.get() for f in status_futures]


def update_issue(issue_url, content):
    match = ISSUE_URL_RE.search(issue_url)
    if not match:
        logger.error(f"update failed: bad issue URL")
        return False

    owner, repo, issue_number = match.groups()
    url = f"https://api.github.com/repos/{owner}/{repo}/issues/{issue_number}"
    github_token = get_github_token()
    headers = {"Accept": "application/vnd.github.v3+json"}
    if github_token:
        headers["Authorization"] = f"token {github_token}"
    data = {"body": content}
    resp = requests.patch(url, json=data, headers=headers)
    if resp.ok:
        return True

    logger.error(f"update failed: {resp.reason}")
    return False


def print_status_table(cci_path, recipes, print_all, jobs):
    status = get_status(cci_path, recipes, jobs)
    table_data = [
        ["Name", "Recipe version", "New version", "Upstream version", "Pending PR"]
    ]
    for s in sorted(status, key=lambda r: r.name):
        if not s.update_possible() and not print_all:
            continue

        if s.deprecated:
            continue

        if s.recipe_version.unknown or s.upstream_version.unknown:
            name_color = fg("red")
        elif s.update_possible():
            name_color = fg("dark_orange")
        else:
            name_color = fg("green")

        pr = s.pr_opened()
        pr_text = "Yes" if pr else "No"

        rv_color = fg("green") if not s.recipe_version.unknown else fg("red")
        uv_color = fg("green") if not s.upstream_version.unknown else fg("red")
        pr_color = fg("green") if pr else fg("dark_orange")

        table_data.append(
            [
                stylize(s.name, name_color),
                stylize(s.recipe_version, rv_color),
                stylize(s.upstream_version.fixed, uv_color),
                stylize(s.upstream_version, uv_color),
                stylize(pr_text, pr_color),
            ]
        )

    table = AsciiTable(table_data)
    print(table.table)
    return 0


def update_status_issue(cci_path, issue_url_list, jobs, dry_run):
    t0 = time.time()
    recipes = get_recipes_list(cci_path)
    status = get_status(cci_path, recipes, jobs)
    updatable = [s for s in status if s.update_possible() and not s.deprecated]
    duration = time.time() - t0

    date = datetime.datetime.now().strftime("%d/%m/%Y")
    text = "\n".join(
        [
            f"* Date: {date}",
            f"* Parsed recipes: {len(recipes)}",
            f"* Updatable recipes: {len(updatable)}",
            f"* Duration: {duration:.1f}s",
            "",
            "Keep in mind this list is auto-generated and the "
            "updatability detection can be flawed.",
            "",
            "You can use conan-center-bot at "
            "https://github.com/qchateau/conan-center-bot "
            "to automatically generate an update.",
            "",
            "|Name|Recipe version|New version|Upstream version|Pending PR|",
            "|----|--------------|-----------|----------------|----------|",
        ]
        + [
            "|".join(
                [
                    "",
                    f"[{s.name}]({s.homepage})" if s.homepage else f"{s.name}",
                    f"{s.recipe_version}",
                    f"{s.upstream_version.fixed}",
                    f"{s.upstream_version}",
                    f"[Yes]({s.pr_opened()['html_url']})" if s.pr_opened() else "No",
                    "",
                ]
            )
            for s in sorted(updatable, key=lambda s: s.name)
        ]
    )

    print(text)
    if not dry_run:
        ok = True
        for issue_url in issue_url_list:
            ok = update_issue(issue_url, text) and ok
    else:
        ok = True

    return 0 if ok else 1



import re

# Global tag blacklist, applied to all projects
TAGS_BLACKLIST = [
    re.compile(r"(.*)[\._-]?(rc|alpha|beta|pre|preview)[\._-]?[0-9]*$", re.IGNORECASE),
    re.compile(r"(.*)test(.*)", re.IGNORECASE),
]

# Project-based blacklist, a whitelist must not be defined
PROJECT_TAGS_BLACKLIST = {
    "libpcap": [re.compile(r"(.*)-bp")],
    "openssl": [re.compile(r"^OpenSSL-fips-(.*)")],
    "libzip": [re.compile(r"^brian-gladman-fcrypt-(.*)")],
    "libselinux": [re.compile(r"^[0-9]{8}(.*)")],
    "pybind11": [re.compile(r"(.*)b[0-9]+$")],
    "lerc": [re.compile(r"^runtimecore_(.*)")],
    "xtensor": [re.compile(r"(.*)-binder[0-9]*$")],
}

# Project-based whitelist, a blacklist must not be defined
PROJECT_TAGS_WHITELIST = {
    "glslang": [re.compile(r"^[0-9]+\.[0-9]+\.[0-9]+$")],
    "coin-clp": [re.compile(r"^releases\/(.*)")],
    "coin-osi": [re.compile(r"^releases\/(.*)")],
    "coin-utils": [re.compile(r"^releases\/(.*)")],
}



import re
import functools

VERSION_DATE_RE = re.compile(r"[0-9]{4}[\.-_]?[0-9]{2}[\.-_]?[0-9]{2}")
VERSION_RE = re.compile(r"[0-9]+(\.[0-9]+)+")
VERSION_DASH_RE = re.compile(r"[0-9]+(-[0-9]+)+")
VERSION_UNDERSCORE_RE = re.compile(r"[0-9]+(_[0-9]+)+")


@functools.total_ordering
class Version:
    UNKNOWN = "unknown"

    def __init__(self, version=UNKNOWN):
        self.original = version
        self.fixed = _fix_version(version)
        self.to_numeric = _to_numeric(self.fixed)
        self.is_date = bool(
            VERSION_DATE_RE.search(
                self.fixed if self.fixed != self.UNKNOWN else self.original
            )
        )

    @property
    def unknown(self):
        return self.original == self.UNKNOWN

    def __hash__(self):
        return hash(self.fixed)

    def __eq__(self, other):
        return self.fixed == other.fixed

    def __lt__(self, other):
        assert isinstance(other, Version)

        if other.to_numeric is None:
            return False
        if self.to_numeric is None:
            return True
        if self.is_date != other.is_date:
            # consider dates as old versions to avoid false positive
            return self.is_date
        return self.to_numeric < other.to_numeric

    def __str__(self):
        return str(self.original)

    def __repr__(self):
        return f"Version<{self.__str__()}>"


def _fix_version(version):
    version = str(version)

    match = VERSION_RE.search(version)
    if match:
        return match.group(0)

    match = VERSION_DASH_RE.search(version)
    if match:
        return match.group(0).replace("-", ".")

    match = VERSION_UNDERSCORE_RE.search(version)
    if match:
        return match.group(0).replace("_", ".")

    return Version.UNKNOWN


def _to_numeric(version):
    try:
        return tuple(int(x) for x in version.split("."))
    except (ValueError, AttributeError):
        return None



import os
import yaml
import typing
import inspect
import logging
import subprocess
import importlib.util
from functools import cached_property, lru_cache

from conans import ConanFile

from .version import Version
from .upstream_project import get_upstream_project
from .cci import cci_interface, LibPullRequest


logger = logging.getLogger(__name__)


def get_recipes_list(cci_path):
    return os.listdir(os.path.join(cci_path, "recipes"))


class RecipeError(RuntimeError):
    pass


class Status(typing.NamedTuple):
    name: str
    homepage: str
    recipe_version: Version
    upstream_version: Version
    deprecated: bool

    def update_possible(self):
        return (
            not self.upstream_version.unknown
            and not self.recipe_version.unknown
            and self.upstream_version.is_date == self.recipe_version.is_date
            and self.upstream_version > self.recipe_version
        )

    def up_to_date(self):
        return (
            not self.upstream_version.unknown
            and not self.recipe_version.unknown
            and self.upstream_version.is_date == self.recipe_version.is_date
            and self.upstream_version <= self.recipe_version
        )

    def pr_opened(self):
        pr = cci_interface.libraries_pull_requests().get(self.name)
        if pr is None:
            return None
        if pr.version == self.upstream_version or pr.version == self.recipe_version:
            return pr.pr
        return None


class Recipe:
    def __init__(self, cci_path, name):
        self.name = name
        self.path = os.path.join(cci_path, "recipes", name)
        self.config_file_path = os.path.join(self.path, "config.yml")

    def config(self):
        if not os.path.exists(self.config_file_path):
            raise RecipeError("No config.yml file")

        with open(self.config_file_path) as fil:
            return yaml.load(fil, Loader=yaml.FullLoader)

    @cached_property
    def upstream(self):
        return get_upstream_project(self)

    @property
    def versions(self):
        return [Version(v) for v in self.config()["versions"].keys()]

    @property
    def versions_folders(self):
        return {Version(k): v["folder"] for k, v in self.config()["versions"].items()}

    @property
    def most_recent_version(self):
        return sorted(self.versions)[-1]

    def status(self, recipe_version=None, upstream_version=None):
        try:
            recipe_version = recipe_version or self.most_recent_version
            recipe_upstream_version = (
                upstream_version or self.upstream.most_recent_version
            )
            homepage = self.upstream.homepage
            deprecated = getattr(
                self.conanfile_class(recipe_version), "deprecated", False
            )
        except RecipeError as exc:
            logger.debug("%s: could not find version: %s", self.name, exc)
            recipe_version = Version()
            recipe_upstream_version = Version()
            homepage = None
            deprecated = None

        return Status(
            self.name,
            homepage,
            recipe_version,
            recipe_upstream_version,
            deprecated,
        )

    @lru_cache
    def conanfile_class(self, version):
        assert isinstance(version, Version)

        version_folder_path = os.path.join(self.path, self.versions_folders[version])

        spec = importlib.util.spec_from_file_location(
            "conanfile", os.path.join(version_folder_path, "conanfile.py")
        )
        conanfile = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(conanfile)

        conanfile_main_class = None
        for symbol_name in dir(conanfile):
            symbol = getattr(conanfile, symbol_name)
            if (
                inspect.isclass(symbol)
                and issubclass(symbol, ConanFile)
                and symbol is not ConanFile
            ):
                conanfile_main_class = symbol
                break

        if conanfile_main_class is None:
            raise RecipeError("Could not find ConanFile class")

        return conanfile_main_class

    def version_exists(self, version):
        return version.fixed in self.config()["versions"]



#!/usr/bin/env python3

import os
import sys
import argparse
import logging

from ccb.recipe import get_recipes_list
from ccb.status import print_status_table, update_status_issue
from ccb.update import update_recipes
from ccb.github import set_github_token

ROOT = os.path.dirname(os.path.realpath(__file__))


def bad_command(args, parser):
    parser.print_usage()
    return 1


def cmd_status(args):
    return print_status_table(
        cci_path=args.cci,
        recipes=args.recipe,
        print_all=args.all,
        jobs=int(args.jobs),
    )


def cmd_update(args):
    return update_recipes(
        cci_path=args.cci,
        recipes=args.recipe,
        choose_version=args.choose_version,
        folder=args.folder,
        run_test=not args.no_test,
        push=args.push,
        force=args.force,
    )


def cmd_update_status_issue(args):
    return update_status_issue(
        cci_path=args.cci,
        issue_url_list=args.issue_url,
        jobs=int(args.jobs),
        dry_run=args.dry_run,
    )


def configure_logging(args):
    if args.verbose > 0 and args.quiet:
        print("--versbose and --quiet cannot be used together")
        sys.exit(1)
    logging.basicConfig(format="%(message)s")
    logger = logging.getLogger("ccb")
    if args.quiet:
        logger.setLevel(logging.ERROR)
    elif args.verbose == 0:
        logger.setLevel(logging.INFO)
    elif args.verbose >= 1:
        logger.setLevel(logging.DEBUG)


def add_subparser(subparsers, name, function, help):
    subparser = subparsers.add_parser(name, help=help)

    subparser.add_argument(
        "--verbose", "-v", action="count", default=0, help="Verbosity level"
    )
    subparser.add_argument("--quiet", "-q", action="store_true")
    subparser.add_argument(
        "--cci",
        help="Path to the conan-center-index repository. Defaults to '../conan-center-index'",
    )
    subparser.add_argument("--github-token", help="Github authentication token")

    subparser.set_defaults(func=function)
    return subparser


def main():
    parser = argparse.ArgumentParser()
    parser.set_defaults(func=lambda args: bad_command(args, parser))
    subparsers = parser.add_subparsers()

    # Status
    parser_status = add_subparser(
        subparsers, "status", cmd_status, help="Display the status of recipes"
    )
    parser_status.add_argument(
        "--all",
        "-a",
        action="store_true",
        help="Display all recipes. By default only updatable recipes are displayed.",
    )
    parser_status.add_argument(
        "--recipe",
        nargs="+",
        help="Restrict the recipes status to the ones given as arguments.",
    )
    parser_status.add_argument(
        "--jobs",
        "-j",
        default=str(10 * os.cpu_count()),
        help="Number of parallel processes.",
    )

    # Update
    parser_update = add_subparser(
        subparsers, "update", cmd_update, help="Auto-update a list of recipes"
    )
    parser_update.add_argument(
        "recipe",
        nargs="*",
        help="List of recipes to update.",
    )
    parser_update.add_argument(
        "--force",
        "-f",
        action="store_true",
        help="Overwrite the branch if it exists, force push if the remote branch exists.",
    )
    parser_update.add_argument(
        "--choose-version",
        action="store_true",
        help="Choose which upstream version to use (defaults to the latest)",
    )
    parser_update.add_argument(
        "--folder",
        help="Choose which recipe folder to use (default to the latest)",
    )
    parser_update.add_argument(
        "--push", action="store_true", help="Push the new branch to origin"
    )
    parser_update.add_argument(
        "--no-test", action="store_true", help="Do not test the updated recipe"
    )

    # Update status issue
    parser_uis = add_subparser(
        subparsers,
        "update-status-issue",
        cmd_update_status_issue,
        help="Update the status issue",
    )
    parser_uis.add_argument("issue_url", nargs="+", help="URL of the issue to update")
    parser_uis.add_argument(
        "--jobs",
        "-j",
        default=str(10 * os.cpu_count()),
        help="Number of parallel processes.",
    )
    parser_uis.add_argument(
        "--dry-run", action="store_true", help="Don't update the issue."
    )

    args = parser.parse_args()

    if args.github_token:
        set_github_token(args.github_token)

    configure_logging(args)

    if not args.cci:
        args.cci = os.path.join(ROOT, "..", "conan-center-index")

    args.cci = os.path.abspath(args.cci)
    if not os.path.exists(args.cci):
        print(f"CCI repository not found at {args.cci}")
        sys.exit(1)

    if not hasattr(args, "recipe") or not args.recipe:
        args.recipe = get_recipes_list(args.cci)

    try:
        sys.exit(args.func(args))
    except KeyboardInterrupt:
        sys.exit(1)


if __name__ == "__main__":
    main()



import re
import abc
import hashlib
import requests
import logging
import tempfile
import subprocess
from functools import cached_property, lru_cache

from .version import Version
from .project_specifics import (
    TAGS_BLACKLIST,
    PROJECT_TAGS_BLACKLIST,
    PROJECT_TAGS_WHITELIST,
)


logger = logging.getLogger(__name__)


class _Unsupported(RuntimeError):
    pass


def get_upstream_project(recipe):
    conanfile_class = recipe.conanfile_class(recipe.most_recent_version)
    for cls in _CLASSES:
        try:
            return cls(recipe, conanfile_class)
        except _Unsupported:
            pass

    logger.debug("%s: unsupported upstream", recipe.name)
    return UnsupportedProject(recipe, conanfile_class)


class UpstreamProject(abc.ABC):
    def __init__(self, recipe, conanfile_class):
        self.recipe = recipe
        self.conanfile_class = conanfile_class

    @property
    def homepage(self) -> str:
        return self.conanfile_class.homepage

    @abc.abstractproperty
    def versions(self) -> dict:
        pass

    @abc.abstractproperty
    def most_recent_version(self) -> Version:
        pass

    @abc.abstractmethod
    def source_url(self, version) -> str:
        pass

    @lru_cache
    def source_sha256_digest(self, version):
        url = self.source_url(version)
        if not url:
            return None
        sha256 = hashlib.sha256()
        tarball = requests.get(url)
        sha256.update(tarball.content)
        return sha256.hexdigest()


class UnsupportedProject(UpstreamProject):
    @cached_property
    def versions(self):
        return {}

    @property
    def most_recent_version(self):
        return Version()

    def source_url(self, version):
        return None


class GitProject(UpstreamProject):
    def __init__(self, recipe, conanfile_class, git_url):
        super().__init__(recipe, conanfile_class)
        self.git_url = git_url
        self.whitelist = PROJECT_TAGS_WHITELIST.get(recipe.name, [])
        self.blacklist = TAGS_BLACKLIST + PROJECT_TAGS_BLACKLIST.get(recipe.name, [])

    @cached_property
    def versions(self):
        logger.debug("%s: fetching tags...", self.recipe.name)
        git_output = subprocess.check_output(
            ["git", "ls-remote", "-t", self.git_url]
        ).decode()

        tags = [ref.replace("refs/tags/", "") for ref in git_output.split()[1::2]]
        tags = list(set(tag[:-3] if tag.endswith("^{}") else tag for tag in tags))
        logger.debug("%s: found tags: %s", self.recipe.name, tags)

        valid_tags = self._filter_tags(tags)
        logger.debug(
            "%s: ignored %s tags", self.recipe.name, len(tags) - len(valid_tags)
        )

        return tuple(Version(tag) for tag in valid_tags)

    @property
    def most_recent_version(self):
        if not self.versions:
            return Version()
        return sorted(self.versions)[-1]

    def _filter_tags(self, tags):
        valid_tags = list()
        if self.whitelist:
            for tag in tags:
                if any(regex.match(tag) for regex in self.whitelist):
                    valid_tags.append(tag)
                else:
                    logger.debug(
                        "%s: tag %s ignored because it does not match any of %s",
                        self.recipe.name,
                        tag,
                        (regex.pattern for regex in self.whitelist),
                    )
        else:
            for tag in tags:
                ignored = False
                for regex in self.blacklist:
                    if regex.match(tag):
                        ignored = True
                        logger.debug(
                            "%s: tag %s ignored because it matches regex %s",
                            self.recipe.name,
                            tag,
                            regex.pattern,
                        )
                        break
                if not ignored:
                    valid_tags.append(tag)
        return valid_tags


class GithubProject(GitProject):
    HOMEPAGE_RE = re.compile(r"https://github.com/([^/]+)/([^/]+)")

    def __init__(self, recipe, conanfile_class):
        match = self.HOMEPAGE_RE.match(conanfile_class.homepage)
        if not match:
            raise _Unsupported()

        owner = match.group(1)
        repo = match.group(2)
        git_url = f"https://github.com/{owner}/{repo}.git"

        super().__init__(recipe, conanfile_class, git_url)
        self.owner = owner
        self.repo = repo

    def source_url(self, version):
        if version.unknown:
            return None
        return f"https://github.com/{self.owner}/{self.repo}/archive/{version.original}.tar.gz"


_CLASSES = [GithubProject]



import tempfile
import subprocess

from .recipe import Recipe


class RecipeInWorktree:
    def __init__(self, recipe):
        self.recipe = recipe
        self.tmpdir = None

    def __enter__(self):
        self.tmpdir = tempfile.TemporaryDirectory()
        try:
            subprocess.check_call(
                [
                    "git",
                    "worktree",
                    "add",
                    "-q",
                    "--checkout",
                    "--detach",
                    self.tmpdir.name,
                    "master",
                ],
                cwd=self.recipe.path,
            )
            return Recipe(self.tmpdir.name, self.recipe.name)
        except:
            self.cleanup()
            raise

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.cleanup()

    def cleanup(self):
        if self.tmpdir is None:
            return

        subprocess.call(
            ["git", "worktree", "remove", "-f", self.tmpdir.name], cwd=self.recipe.path
        )
        self.tmpdir.cleanup()



import re
import typing
import logging
import requests
import functools
from .github import get_github_token
from .version import Version


PR_REGEX = re.compile(r"Specify library name and version:[\s\*]*(.*)/([^\s]+)")


logger = logging.getLogger(__name__)


class LibPullRequest(typing.NamedTuple):
    library: str
    version: Version
    pr: typing.Any


class _CciInterface:
    def __init__(self):
        self.owner = "conan-io"
        self.repo = "conan-center-index"

    @functools.lru_cache
    def pull_requests(self):
        github_token = get_github_token()
        headers = {"Accept": "application/vnd.github.v3+json"}
        if github_token:
            headers["Authorization"] = f"token {github_token}"
        url = f"https://api.github.com/repos/{self.owner}/{self.repo}/pulls"

        prs = list()

        page = 1
        while True:
            logger.debug("getting PR page %s", page)
            params = {"page": str(page), "per_page": "100"}
            res = requests.get(url, params=params, headers=headers)
            res.raise_for_status()
            results = res.json()
            prs.extend(results)
            page += 1

            logger.debug("%s results", len(results))
            if not results:
                break

        return prs

    @functools.lru_cache
    def libraries_pull_requests(self):
        prs = self.pull_requests()
        lib_prs = dict()
        for pr in prs:
            match = PR_REGEX.search(pr.get("body", ""))
            if not match:
                continue
            name = match.group(1)
            version = Version(match.group(2))
            lib_prs[name] = LibPullRequest(name, version, pr)
        return lib_prs


cci_interface = _CciInterface()



import os
from setuptools import setup


def get_requires(filename):
    requirements = []
    with open(filename, "rt") as req_file:
        for line in req_file.read().splitlines():
            if not line.strip().startswith("#"):
                requirements.append(line)
    return requirements


setup(
    name="conan-center-bot",
    version=os.environ.get("CCB_VERSION", "0.0.0"),
    license="GPLv3",
    packages=["ccb"],
    entry_points={
        "console_scripts": [
            "conan-center-bot=ccb.__main__:main",
        ],
    },
    include_package_data=True,
    install_requires=get_requires("requirements.txt"),
    author="Quentin Chateau",
    author_email="quentin.chateau@gmail.com",
    description="A bot to automatically update conan-center-index",
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/qchateau/conan-center-bot",
    classifiers=[
        "Development Status :: 1 - Planning",
        "Programming Language :: Python :: 3",
        "Programming Language :: C",
        "Programming Language :: C++",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
)
